"""Define a SmartCocoon Fan class."""

import asyncio
import logging
import uuid
from datetime import datetime
from typing import Any, Callable, Optional

import paho.mqtt.client as mqtt

from pysmartcocoon.api import SmartCocoonAPI
from pysmartcocoon.const import (
    API_FANS_URL,
    API_URL,
    DEFAULT_FAN_POWER_PCT,
    MQTT_BROKER,
    MQTT_KEEPALIVE,
    MQTT_PORT,
    EntityType,
    FanMode,
)

_LOGGER: logging.Logger = logging.getLogger(__name__)


class Fan:
    """Define the fan."""

    def __init__(
        self,
        fan_id: str,
        api: SmartCocoonAPI,
        async_update_fan_callback: Optional[Callable[[str, Any], None]] = None,
    ) -> None:
        """Initialize."""

        # Fan attributes from SmartCocoon
        self._fan_id: int = fan_id
        self._id: int = None
        self._fan_on: bool = None
        self._firmware_version: str = None
        self._is_room_estimating: bool = None
        self._connected: bool = None
        self._last_connection: datetime = None
        self._mode: str = None
        self._power: int = None
        self._predicted_room_temperature: float = None
        self._room_id: int = None
        self._thermostat_vendor: int = None
        self._mqtt_username: str = None
        self._mqtt_password: str = None

        # Extra attributes not provided by API
        self._room_name: str = None

        self._api = api

        # MQTT attributes
        self._async_update_fan_callback = async_update_fan_callback
        self._mqtt_connected = False
        self._mqttc: mqtt.Client = None
        self._loop = asyncio.get_running_loop()

    @property
    def id(self) -> str:
        """Return Fan id.

        This is a numerical ID generated by SmartCocoon
        This value can change if the fan is re-added to
        an cloud account
        """
        return self._id

    @property
    def fan_id(self) -> str:
        """Return Fan fan_id.

        This is the physical ID printed on the fan
        """
        return self._fan_id

    @property
    def fan_on(self) -> str:
        """Return a bool indicating if the fan is on or off
        On = True
        Off = False
        """
        return self._fan_on

    @property
    def firmware_version(self) -> str:
        """Return the firmware version of the fan"""
        return self._firmware_version

    @property
    def is_room_estimating(self) -> bool:
        """Return value to indicate if the room is estimating temp
        Estimating = True
        Not Estimating = False
        """
        return self._is_room_estimating

    @property
    def connected(self) -> bool:
        """Return a bool indicating if the fan is connected
        Connected = True
        Not Connected = False
        """
        return self._connected

    @property
    def last_connection(self) -> datetime:
        """Return a bool indicating if the fan is connected
        Connected = True
        Not Connected = False
        """
        return self._last_connection

    @property
    def mode(self) -> str:
        """Return the current mode setting of the fan"""
        return self._mode

    @property
    def mode_enum(self) -> FanMode:
        """Return the current mode setting of the fan"""
        return FanMode(self._mode)

    @property
    def power(self) -> int:
        """Return the current power setting of the fan
        % value is multiplied by 100
        100% = power level of 10000
        """
        return self._power

    @property
    def speed_pct(self) -> int:
        """Return the current power setting of the fan
        by %
        """
        return self._power / 100

    @property
    def predicted_room_temperature(self) -> float:
        """Return the predicted room temperature of the fan"""
        return self._predicted_room_temperature

    @property
    def room_name(self) -> str:
        """Return the room name of the fan"""
        return self._room_name

    def set_speed_pct(self, fan_speed_pct: int) -> bool:
        """Update the power of the fan"""

        if fan_speed_pct > 100:
            _LOGGER.debug(
                "Fan ID: %s - Fan speed of %s%% is invalid, must be "
                "between 0%% and 100%%",
                self.fan_id,
                str(fan_speed_pct),
            )

        _LOGGER.debug(
            "Fan ID: %s - Updating fan speed to %s%%", self.fan_id, str(fan_speed_pct)
        )

        self._power = fan_speed_pct * 100
        return True

    def set_room_name(self, room_name: str) -> bool:
        """Update the room_name"""

        _LOGGER.debug("Fan ID: %s - Updating room_name to '%s'", self.fan_id, room_name)

        self._room_name = room_name
        return True

    @property
    def room_id(self) -> str:
        """Return the room id of the fan"""
        return self._room_id

    @property
    def thermostat_vendor(self) -> str:
        """Return the thermostat vendor of the fan"""
        return self._thermostat_vendor

    @property
    def mqtt_username(self) -> str:
        """Return the MQTT username of the fan"""
        return self._mqtt_username

    @property
    def mqtt_password(self) -> str:
        """Return the MQTT password of the fan"""
        return self._mqtt_password

    async def async_set_fan_modes(
        self, fan_mode: FanMode = None, fan_speed_pct: int = None
    ) -> bool:
        """Set the fan mode and speed."""

        _LOGGER.debug(
            "Fan ID: %s - In async_set_fan_mode with fan_mode: "
            "%s and fan_speed: %s%%",
            self.fan_id,
            fan_mode,
            str(fan_speed_pct),
        )

        if fan_mode is None and fan_speed_pct is None:
            _LOGGER.debug(
                "async_set_fan_modes must provide a value for "
                "fan_mode and/or fan_speed"
            )
            return False

        if fan_mode is None:
            # Determine value for fan_mode
            if fan_speed_pct == 0:
                if self.mode_enum is FanMode.ON:
                    fan_mode = FanMode.OFF
                else:
                    fan_mode = self.mode_enum
            else:
                if self.mode_enum is FanMode.OFF:
                    fan_mode = FanMode.ON
                else:
                    fan_mode = self.mode_enum

        # Update fan mode if changed
        if self.mode_enum != fan_mode:
            self._mode = fan_mode.value

        if fan_speed_pct is None:
            if fan_mode is FanMode.OFF:
                fan_speed_pct = self.speed_pct
            elif fan_mode is FanMode.ON and self.speed_pct == 0:
                fan_speed_pct = DEFAULT_FAN_POWER_PCT
            else:
                fan_speed_pct = self.speed_pct
        else:
            if fan_speed_pct > 100:
                _LOGGER.debug(
                    "Fan ID: %s - Fan speed of %s%% is invalid, "
                    "must be between 0%% and 100%%",
                    self.fan_id,
                    str(fan_speed_pct),
                )
                return False

        # Update power if changed
        if self.speed_pct != fan_speed_pct:
            self.set_speed_pct(fan_speed_pct)

        await self._async_set_fan(fan_mode)

    async def _async_set_fan(self, fan_mode: FanMode = None) -> bool:
        """Call the API to update the fan mode and speed."""

        request_body = {}
        request_body.setdefault("json", {})
        request_body["json"]["mode"] = self.mode
        request_body["json"]["power"] = self.power

        await self._api.async_request("PUT", f"{API_FANS_URL}{self.id}", **request_body)

        _LOGGER.debug(
            "Fan ID: %s - Fan Mode was set to %s, speed to %s",
            self.fan_id,
            self.mode,
            self.speed_pct,
        )

        await self._async_update_fan()

        if fan_mode == FanMode.ON.value and not self.fan_on:
            _LOGGER.debug("Fan ID: %s - Changing fan_on to 'True'", self.fan_id)
            self._fan_on = True
        elif fan_mode == FanMode.OFF.value:
            _LOGGER.debug("Fan ID: %s - Changing fan_on to 'False'", self.fan_id)
            self._fan_on = False

    async def async_update_api_data(
        self,
        data: dict[str, Any],
    ) -> bool:
        """Selectively update the fan attributes with API data"""

        _LOGGER.debug("Fan ID: %s - In async_update_api_data", data["fan_id"])

        self._id: int = data["id"]
        self._fan_id: str = data["fan_id"]

        # Fan attributes from SmartCocoon
        self._mode: str = data["mode"]

        # fan_on does not always reflect the current mode, mode is more
        # accurate if set to always_on or always_off

        if self._mode == FanMode.ON.value:
            self._fan_on = True
        elif self._mode == FanMode.OFF.value:
            self._fan_on = False
        else:
            self._fan_on: bool = data["fan_on"]

        self._firmware_version = data["firmware_version"]
        self._is_room_estimating = data["is_room_estimating"]
        self._connected = data["connected"]
        self._last_connection = data["last_connection"]
        self._power = data["power"]
        self._predicted_room_temperature = data["predicted_room_temperature"]
        self._room_id = data["room_id"]
        self._thermostat_vendor = data["thermostat_vendor"]
        self._mqtt_username = data["mqtt_username"]
        self._mqtt_password = data["mqtt_password"]

    async def async_set_async_update_fan_callback(
        self, async_update_fan_callback: Callable[[str, Any], None]
    ) -> None:

        self._async_update_fan_callback = async_update_fan_callback

    async def _async_update_fan(self) -> bool:

        _LOGGER.debug("Fan ID: %s - Updating fan attributes from cloud", self.fan_id)

        entity = EntityType.FANS.value
        response = await self._api.async_request("GET", f"{API_URL}{entity}/{self._id}")

        if len(response) is not None:
            await self.async_update_api_data(response)

            if self._async_update_fan_callback:
                _LOGGER.debug(
                    "Fan ID: %s - Executing callback for fan update", self.fan_id
                )
                await self._async_update_fan_callback()
                _LOGGER.debug(
                    "Fan ID: %s - Completed callback for fan update", self.fan_id
                )

        return True

    def _mqtt_on_connect(self, _mqttc, userdata, flags, rc: int):
        if rc == 0:
            _LOGGER.debug("Fan ID: %s - MQTT connection successful", self.fan_id)
            self._mqtt_connected = True
        else:
            _LOGGER.debug("Fan ID: %s - MQTT connection failed %i", self.fan_id, rc)
            self._mqtt_connected = False

    def _mqtt_on_disconnect(self, _mqttc, userdata, rc: int):
        if rc == 0:
            _LOGGER.debug(
                "Fan ID: %s - MQTT has been successfully" "disconnected", self.fan_id
            )
        else:
            _LOGGER.debug(
                "Fan ID: %s - MQTT unexpected disconnect: %i", self.fan_id, rc
            )
            asyncio.run_coroutine_threadsafe(self._async_start_mqtt(), self._loop)

        self._mqtt_connected = False

    def _mqtt_on_message_status(self, _mqttc, userdata, message: mqtt.MQTTMessage):
        payload = str(message.payload)
        _LOGGER.debug("Fan ID: %s - MQTT message received: %s", self.fan_id, payload)

        # Format of topic should be "nnnn_fan_id/status"
        # Where nnnn = location_id

        # Ignore if payload contains 'sendKeepAlive'
        if payload.find("sendKeepAlive"):
            asyncio.run_coroutine_threadsafe(self._async_update_fan(), self._loop)

    async def _async_start_mqtt(self) -> bool:
        """Start MQTT subscriptions."""

        _LOGGER.debug("Fan ID: %s - In async_start_mqtt", self.fan_id)

        if self.mqtt_username is None:
            _LOGGER.debug("Fan ID: %s - MQTT username is not provided", self.fan_id)
            return False

        client_id = mqtt.base62(uuid.uuid4().int, padding=22)
        self._mqttc = mqtt.Client(client_id, protocol=mqtt.MQTTv311)
        self._mqttc.username_pw_set(self.mqtt_username, password=self.mqtt_password)
        self._mqttc.on_connect = self._mqtt_on_connect
        self._mqttc.connect(MQTT_BROKER, port=MQTT_PORT, keepalive=MQTT_KEEPALIVE)
        self._mqttc.on_disconnect = self._mqtt_on_disconnect
        self._mqttc.message_callback_add(
            f"{self._mqtt_username}/status", self._mqtt_on_message_status
        )
        self._mqttc.subscribe(f"{self._mqtt_username}/status")
        self._mqttc.loop_start()
        return True

    async def _async_stop_mqtt(self) -> bool:
        """Stop MQTT subscriptions."""

        _LOGGER.debug("Fan ID: %s - Stopping MQTT", self.fan_id)

        self._mqttc.disconnect()
        self._mqttc.loop_stop()

    async def async_start_services(self) -> bool:
        """Start all services."""

        _LOGGER.debug("Fan ID: %s - Starting services", self.fan_id)
        await self._async_start_mqtt()

    async def async_stop_services(self) -> bool:
        """Stop all services."""

        _LOGGER.debug("Fan ID: %s - Stopping services", self.fan_id)
        await self._async_stop_mqtt()



# coding: utf-8
"""Tests SmartCocoon api module."""
import asyncio

import aiohttp 
import time
import logging

import pytest

from pysmartcocoon.manager import SmartCocoonManager
from pysmartcocoon.fan import Fan
from pysmartcocoon.const import (
    FanMode,
)

# Enter correct real values here for the tests to complete successfully with real SmartCocoon Server calls.
USERNAME = "dpmn.home@gmail.com"
PASSWORD = "JOY-refer-percept"
FAN_ID = "6fcf50"  # This is the physical fan ID printed from the fan

logging.basicConfig(level=logging.DEBUG)
_LOGGER: logging.Logger = logging.getLogger(__name__)

@pytest.mark.skip("Not an automated test but an example of usage with real values.")


async def test_integration_fan_control() -> None:
    """Test authentification then control fan."""

    async with aiohttp.ClientSession() as session:
        # Init manager
        manager = SmartCocoonManager(
            session=session
        )

        if not await manager.async_start_services( 
            username = USERNAME, 
            password = PASSWORD,
            use_mqtt = False 
        ):
            _LOGGER.debug("Authentication failed")
            return

        # Test fan controls
        WAIT_FOR = 5

        await manager.async_fan_turn_on(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_fan_turn_off(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_speed(FAN_ID, 0)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_modes(
            fan_id=FAN_ID,
            fan_mode = FanMode.ON, 
            fan_speed_pct = 100)        
        time.sleep(WAIT_FOR)
        await manager.async_fan_turn_off(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_fan_turn_on(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_speed(FAN_ID, 0)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_modes(
            fan_id=FAN_ID,
            fan_mode = FanMode.ON, 
            fan_speed_pct = 33)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_speed(FAN_ID, 33)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_speed(FAN_ID, 101)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_auto(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_eco(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_set_fan_modes(
            fan_id=FAN_ID,
            fan_mode = None, 
            fan_speed_pct = None)
        time.sleep(WAIT_FOR)
        await manager.async_fan_turn_off(FAN_ID)
        time.sleep(WAIT_FOR)

        #await manager.async_stop_services()
        logging.info("Successfully shutdown the service.")        

asyncio.run(test_integration_fan_control())


"""Define a manager to interact with SmartCocoon"""
import asyncio
import logging
from typing import Optional

from aiohttp import ClientSession

from pysmartcocoon.api import SmartCocoonAPI
from pysmartcocoon.const import API_URL, DEFAULT_TIMEOUT, EntityType, FanMode
from pysmartcocoon.fan import Fan
from pysmartcocoon.location import Location
from pysmartcocoon.room import Room
from pysmartcocoon.thermostat import Thermostat

_LOGGER: logging.Logger = logging.getLogger(__name__)


class SmartCocoonManager:
    """Define the main controller class to communicate with the
    SmartCocoon cloud API
    """

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        request_timeout: int = DEFAULT_TIMEOUT,
    ) -> None:

        self._api = SmartCocoonAPI(session, request_timeout)

        self._api_connected: bool = False

        """Variables to store SmartCocoon response data"""
        self._locations: dict[int, Location] = {}
        self._thermostats: dict[int, Thermostat] = {}
        self._rooms: dict[int, Room] = {}
        self._fans: dict[int, Fan] = {}

    @property
    def locations(self):
        """Return list of Locations."""
        return self._locations

    @property
    def thermostats(self):
        """Return list of Thermostats."""
        return self._thermostats

    @property
    def rooms(self):
        """Return list of Rooms."""
        return self._rooms

    @property
    def fans(self):
        """Return list of Fans."""
        return self._fans

    async def async_start_services(
        self, username: str, password: str, use_mqtt: bool = True
    ) -> bool:

        _LOGGER.debug("Starting services")

        # Authenticate with the API
        self._api_connected = await self._api.async_authenticate(username, password)

        if self._api_connected:
            # Make API calls to initial data
            await self.async_update_data()

        # Start Fan services if enabled
        self._use_mqtt = use_mqtt
        if self._use_mqtt:
            for fan_id in self._fans:
                _LOGGER.debug("Starting services for fan: %s", fan_id)
                await self._fans[fan_id].async_start_services()
        else:
            _LOGGER.debug("Fan services have been disabled")

        return self._api_connected

    async def async_stop_services(self) -> bool:

        _LOGGER.debug("Stopping services")

        # Close the MQTT sessions managed by the fans if enabled

        if self._use_mqtt:
            for fan_id in self._fans:
                _LOGGER.debug("Stopping services for fan: %s", fan_id)
                await self._fans[fan_id].async_stop_services()

        return True

    async def async_update_data(self) -> None:
        tasks = []
        tasks.append(self.async_update_locations())
        tasks.append(self.async_update_thermostats())
        tasks.append(self.async_update_rooms())
        await asyncio.gather(*tasks)

        """Fans requires rooms to be loaded in order to add the
           room_name to the fan
        """
        await self.async_update_fans()

    async def async_update_locations(self) -> dict[int, Location]:

        entity = EntityType.LOCATIONS.value
        response = await self._api.async_request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                location = Location(data=item)
                self._locations[location.id] = location

        return self._locations

    async def async_update_thermostats(self) -> dict[int, Thermostat]:

        entity = EntityType.THERMOSTATS.value
        response = await self._api.async_request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                thermostat = Thermostat(data=item)
                self._thermostats[thermostat.id] = thermostat

        return self._thermostats

    async def async_update_rooms(self) -> dict[int, Room]:

        entity = EntityType.ROOMS.value
        response = await self._api.async_request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                room = Room(data=item)
                self._rooms[room.id] = room

        return self._rooms

    async def async_update_fans(self) -> dict[int, Fan]:

        entity = EntityType.FANS.value
        response = await self._api.async_request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for data in response[entity]:
                # Check to see if Fan exists in _fans
                fan_id = data["fan_id"]

                if fan_id not in self._fans:
                    fan = Fan(fan_id=fan_id, api=self._api)
                    self._fans[fan_id] = fan

                await self._fans[fan_id].async_update_api_data(data)
                room_id = self._fans[fan_id].room_id
                room_name = await self.async_get_room_name(room_id)
                self._fans[fan_id].set_room_name(room_name)

        return self._fans

    async def async_get_room_name(self, room_id: int) -> str:

        if room_id in self._rooms:
            room_name = self.rooms[room_id].name

            _LOGGER.debug(
                "In async_get_room_name for room_id: %s, " "found room nane: %s",
                room_id,
                room_name,
            )
        else:
            room_name = None
            _LOGGER.debug("In async_get_room_name, room_id: %s was not nout", room_id)

        return room_name

    async def async_fan_turn_on(self, fan_id: str) -> None:
        """Turn on fan."""

        await self._fans[fan_id].async_set_fan_modes(fan_mode=FanMode.ON)

    async def async_fan_turn_off(self, fan_id: str) -> None:
        """Turn on fan."""

        await self._fans[fan_id].async_set_fan_modes(fan_mode=FanMode.OFF)

    async def async_set_fan_auto(self, fan_id: str) -> None:
        """Enable auto mode on fan."""

        await self._fans[fan_id].async_set_fan_modes(fan_mode=FanMode.AUTO)

    async def async_set_fan_eco(self, fan_id: str) -> None:
        """Enable eco mode on fan."""

        await self._fans[fan_id].async_set_fan_modes(fan_mode=FanMode.ECO)

    async def async_set_fan_modes(
        self, fan_id: str, fan_mode: FanMode, fan_speed_pct: int
    ) -> None:
        """Set fan mode and speed."""

        await self._fans[fan_id].async_set_fan_modes(
            fan_mode=fan_mode, fan_speed_pct=fan_speed_pct
        )

    async def async_set_fan_speed(self, fan_id: str, fan_speed_pct: int) -> None:
        """Set fan speed."""

        await self._fans[fan_id].async_set_fan_modes(fan_speed_pct=fan_speed_pct)



"""Define a manager to interact with SmartCocoon"""
import asyncio
import logging
import traceback
from datetime import datetime, timedelta
from typing import Any, Optional, cast

import async_timeout
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from pysmartcocoon.const import API_AUTH_URL, API_HEADERS, DEFAULT_TIMEOUT

_LOGGER: logging.Logger = logging.getLogger(__name__)


class SmartCocoonAPI:
    """This class will communicate with the SmartCocoon cloud API"""

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        request_timeout: int = DEFAULT_TIMEOUT,
    ) -> None:

        self._session = session
        self._request_timeout = request_timeout

        self._headersAuth = API_HEADERS
        self._authenticated = False

    async def async_authenticate(self, username: str, password: str) -> bool:

        self._authenticated = False

        # Authenticate with user and pass
        request_body = {}
        request_body.setdefault("json", {})
        request_body["json"]["email"] = username
        request_body["json"]["password"] = password

        await self.async_request("POST", API_AUTH_URL, **request_body)

        return self._authenticated

    async def async_request(self, method: str, url: str, **kwargs) -> dict:
        """Make a request using token authentication.
        Args:
            method: Method for the HTTP request (example "GET" or "POST").
            path: path of the REST API endpoint.
        Returns:
            the Response object corresponding to the result of the API request.
        """

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            timeout_value = ClientTimeout(total=DEFAULT_TIMEOUT)
            session = ClientSession(timeout=timeout_value)

        assert session

        data = None

        _LOGGER.debug("Calling SmartCocoon API - method: %s, url: %s", method, url)

        try:
            async with async_timeout.timeout(self._request_timeout):
                response = await session.request(
                    method, url, ssl=False, headers=self._headersAuth, **kwargs
                )
                _LOGGER.debug("SmartCocoon API response status: %s", response.status)
                response.raise_for_status()
                data = await response.json(content_type=None)
        except ClientError as err:
            if "401" in str(err):
                # Authentication failed
                return None
            elif "403" in str(err):
                # Forbidden error - likely needs to re-authenticate
                return None
            elif "Server disconnected" in str(err):
                # Server disconnected error
                _LOGGER.error("SmartCocoon API response error: %s", err.message)
                return None
        except asyncio.TimeoutError:
            _LOGGER.error("API call to SmartCocoon timed out")
            return None
        except Exception:
            _LOGGER.error("API call to SmartCocoon failed with expected error")
            _LOGGER.error(traceback.format_exc())
            return None
        finally:
            if not use_running_session:
                await session.close()

        # If this request is for authorization, save auth data
        if url == API_AUTH_URL:
            self._bearerToken: str = response.headers["access-token"]
            self._bearerTokenExpiration: datetime = datetime.now() + timedelta(
                seconds=int(response.headers["expiry"]) - 10
            )
            self._apiClient: str = response.headers["client"]

            self._headersAuth["access-token"] = self._bearerToken
            self._headersAuth["client"] = self._apiClient
            self._headersAuth["uid"] = data["data"]["email"]

            self._user_id: str = data["data"]["id"]
            self._authenticated = True

        if data is not None:
            return cast(dict[str, Any], data)
        else:
            _LOGGER.error("Response data is None")
            return



"""Define package errors."""


class SmartCocoonError(Exception):
    """Define a base error."""

    pass


class UnauthorizedError(SmartCocoonError):
    """Define an error related to invalid requests."""

    pass


class RequestError(SmartCocoonError):
    """Define an error related to invalid requests."""

    pass


class TokenExpiredError(SmartCocoonError):
    """Define an error for expired access tokens that can't be refreshed."""

    pass


ERROR_CODES = {1: "The email has not been validated"}



# SmartCocoon constants

from enum import Enum

# API Data
API_URL = "https://app.mysmartcocoon.com/api/"
API_AUTH_URL = API_URL + "auth/sign_in"
API_FANS_URL = API_URL + "fans/"
API_HEADERS = {
    "Content-Type": "application/json",
    "User-Agent": "SmartCocoon/1 CFNetwork/1312 Darwin/21.0.0",
}

# Default fan speed is 38% in the SmartCocoon app
# but 33% works better for HomeAssistant
DEFAULT_FAN_POWER_PCT = 33

# MQTT Data
MQTT_BROKER = "nlb-production-04b5ef4727e9be7d.elb.us-east-2.amazonaws.com"
MQTT_PORT = 1883
MQTT_KEEPALIVE = 600

DEFAULT_TIMEOUT: int = 30


class EntityType(Enum):
    # SmartCocoon entity types

    LOCATIONS = "client_systems"
    THERMOSTATS = "thermostats"
    ROOMS = "rooms"
    FANS = "fans"


class FanMode(Enum):
    """Fan mode."""

    ECO = "eco"
    OFF = "always_off"
    ON = "always_on"
    AUTO = "auto"


class FanState(Enum):
    """Fan State."""

    FAN_OFF = "false"
    FAN_ON = "true"



"""Define a SmartCocoon Thermostat class."""
# pylint: disable=too-few-public-methods,too-many-instance-attributes

from typing import Any


class Thermostat:  # pylint: disable=too-many-instance-attributes
    """Define the thermostat."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize."""
        self.id: int = data["id"]
        self.name: str = data["name"]
        self.thermostat_id: id = data["thermostat_id"]
        self.token: str = data["token"]
        self.hvac_mode: str = data["hvac_mode"]
        self.hvac_state: str = data["hvac_state"]
        self.temperature: float = data["temperature"]
        self.target_temperature: float = data["target_temperature"]
        self.vendor: str = data["vendor"]



"""Define a client to interact with SmartCocoon"""
import asyncio
import logging
from datetime import datetime, timedelta
from typing import Any, Optional, cast

import async_timeout
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from pysmartcocoon.const import (
    API_AUTH_URL,
    API_FANS_URL,
    API_HEADERS,
    API_URL,
    DEFAULT_TIMEOUT,
    EntityType,
    FanMode,
)
from pysmartcocoon.fan import Fan
from pysmartcocoon.location import Location
from pysmartcocoon.room import Room
from pysmartcocoon.thermostat import Thermostat

_LOGGER: logging.Logger = logging.getLogger(__name__)


class Client:
    """Define the main controller class to communicate with the SmartCocoon cloud API"""

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        request_timeout: int = DEFAULT_TIMEOUT,
    ) -> None:

        self._session = session
        self._request_timeout = request_timeout

        self._headersAuth = API_HEADERS
        self._authenticated = False

        """Variables to store SmartCocoon response data"""
        self.locations: dict[int, Location] = {}
        self.thermostats: dict[int, Thermostat] = {}
        self.rooms: dict[int, Room] = {}
        self.fans: dict[int, Fan] = {}

    async def authenticate(self, username: str, password: str) -> bool:

        self._authenticated = False

        # Authenticate with user and pass
        request_body = {}
        request_body.setdefault("json", {})
        request_body["json"]["email"] = username
        request_body["json"]["password"] = password

        await self._request("POST", API_AUTH_URL, **request_body)

        return self._authenticated

    async def _request(self, method: str, url: str, **kwargs) -> dict:
        """Make a request using token authentication.
        Args:
            method: Method for the HTTP request (example "GET" or "POST").
            path: path of the REST API endpoint.
        Returns:
            the Response object corresponding to the result of the API request.
        """

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        assert session

        try:
            async with async_timeout.timeout(self._request_timeout):
                async with session.request(
                    method, url, ssl=False, headers=self._headersAuth, **kwargs
                ) as response:
                    response.raise_for_status()
                    data = await response.json(content_type=None)
        except ClientError as err:
            if "401" in str(err):
                # Authentication failed
                return None
        except asyncio.TimeoutError:
            _LOGGER.error("API call to SmartCocoon timed out")
            return None
        finally:
            if not use_running_session:
                await session.close()

        # If this request is for authorization, save auth data
        if url == API_AUTH_URL:
            self._bearerToken: str = response.headers["access-token"]
            self._bearerTokenExpiration: datetime = datetime.now() + timedelta(
                seconds=int(response.headers["expiry"]) - 10
            )
            self._apiClient: str = response.headers["client"]

            self._headersAuth["access-token"] = self._bearerToken
            self._headersAuth["client"] = self._apiClient
            self._headersAuth["uid"] = data["data"]["email"]

            self._user_id: str = data["data"]["id"]
            self._authenticated = True

        return cast(dict[str, Any], data)

    async def load_data(self) -> None:
        tasks = []
        tasks.append(self._load_locations())
        tasks.append(self._load_thermostats())
        tasks.append(self._load_rooms())
        tasks.append(self._load_fans())

        await asyncio.gather(*tasks)

    async def _load_locations(self) -> dict[int, Location]:
        # Init locations
        self.locations: dict[int, Location] = {}

        entity = EntityType.LOCATIONS.value
        response = await self._request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                location = Location(data=item)
                self.locations[location.id] = location

        return self.locations

    async def _load_thermostats(self) -> dict[int, Thermostat]:

        # Init thermostats
        self.thermostats: dict[int, Thermostat] = {}

        entity = EntityType.THERMOSTATS.value
        response = await self._request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                thermostat = Thermostat(data=item)
                self.thermostats[thermostat.id] = thermostat

        return self.thermostats

    async def _load_rooms(self) -> dict[int, Room]:

        # Init rooms
        self.rooms: dict[int, Room] = {}

        entity = EntityType.ROOMS.value
        response = await self._request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                room = Room(data=item)
                self.rooms[room.id] = room

        return self.rooms

    async def _load_fans(self) -> dict[int, Fan]:

        # Init fans
        self.fans: dict[int, Fan] = {}

        entity = EntityType.FANS.value
        response = await self._request("GET", f"{API_URL}{entity}")

        if len(response) != 0:
            for item in response[entity]:
                fan = Fan(data=item)
                self.fans[fan.fan_id] = fan

        return self.fans

    async def _set_fan_mode(self, fan_id: str, fan_mode: FanMode) -> None:
        """Set the fan mode."""

        request_body = {}
        request_body.setdefault("json", {})
        request_body["json"]["mode"] = fan_mode.value

        fan = self.fans[fan_id]
        await self._request("PUT", f"{API_FANS_URL}{fan.id}", **request_body)

        if fan_mode == FanMode.ON:
            fan.fan_on = True
        elif fan_mode == FanMode.OFF:
            fan.fan_on = False

    async def fan_turn_on(self, fan_id: str) -> None:
        """Turn on fan."""
        await self._set_fan_mode(fan_id, FanMode.ON)

        _LOGGER.debug("Fan %s was turned on", fan_id)

    async def fan_turn_off(self, fan_id: str) -> None:
        """Turn on fan."""
        await self._set_fan_mode(fan_id, FanMode.OFF)

        _LOGGER.debug("Fan %s was turned off", fan_id)

    async def fan_auto(self, fan_id: str) -> None:
        """Enable auto mode on fan."""
        await self._set_fan_mode(fan_id, FanMode.AUTO)

        _LOGGER.debug("Fan %s was set to auto", fan_id)

    async def fan_eco(self, fan_id: str) -> None:
        """Enable eco mode on fan."""
        await self._set_fan_mode(fan_id, FanMode.ECO)

        _LOGGER.debug("Fan %s was set to eco", fan_id)

    async def fan_set_speed(self, fan_id: str, fan_speed: int) -> None:
        """Set fan speed."""

        if fan_speed > 100:
            _LOGGER.debug(
                "Fan speed of %s is invalid, must be between 0 and 100", fan_speed
            )
            return

        request_body = {}
        request_body.setdefault("json", {})
        request_body["json"]["power"] = fan_speed * 100

        fan = self.fans[fan_id]
        await self._request("PUT", f"{API_FANS_URL}{fan.id}", **request_body)

        _LOGGER.debug("Fan %s speed was set to %s%", fan_id, fan_speed)



"""Tests SmartCocoon api module."""
import asyncio
import logging
import time

import aiohttp
import pytest

from pysmartcocoon.fan import Fan
from pysmartcocoon.manager import SmartCocoonManager

# Enter correct real values here for the tests to complete successfully with real SmartCocoon Server calls.
USERNAME = "user@domain.com"
PASSWORD = "password"
FAN_ID = "1abc23"  # This is the physical fan ID printed from the fan

logging.basicConfig(level=logging.DEBUG)
_LOGGER: logging.Logger = logging.getLogger(__name__)


@pytest.mark.skip("Not an automated test but an example of usage with real values.")
def async_fan_update_callback(fan: Fan) -> None:
    """Handle fan updates."""
    _LOGGER.debug(
        f"Received an update for Fan {fan.fan_id}: fan_on = {fan.fan_on}, power: {fan.power}, mode: {fan.mode}"
    )


async def test_integration_fan_control() -> None:
    """Test authentification then control fan."""

    async with aiohttp.ClientSession() as session:
        # Init manager
        manager = SmartCocoonManager(session=session)

        if not await manager.async_start_services(USERNAME, PASSWORD):
            _LOGGER.debug("Authentication failed")
            return

        # Test fan controls
        WAIT_FOR = 5

        await manager.async_fan_turn_on(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_fan_set_speed(FAN_ID, 50)
        time.sleep(WAIT_FOR)
        await manager.async_fan_set_speed(FAN_ID, 100)
        time.sleep(WAIT_FOR)
        await manager.async_fan_auto(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_fan_eco(FAN_ID)
        time.sleep(WAIT_FOR)
        await manager.async_fan_turn_off(FAN_ID)
        time.sleep(WAIT_FOR)

        await manager.async_stop_services()


asyncio.run(test_integration_fan_control())



"""Define a SmartCocoon Room class."""
# pylint: disable=too-few-public-methods,too-many-instance-attributes

from typing import Any


class Room:  # pylint: disable=too-many-instance-attributes
    """Define the room."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize."""
        self._id: int = data["id"]
        self._name: str = data["name"]
        self._desired_temperature: float = data["desired_temperature"]
        self._hvac_mode: str = data["hvac_mode"]
        self._hvac_state: str = data["hvac_state"]
        self._is_estimating: bool = data["is_estimating"]
        self._predicted_temperature: float = data["predicted_temperature"]
        self._target_temperature: float = data["target_temperature"]
        self._temperature: float = data["temperature"]
        self._thermostat_id: int = data["thermostat_id"]

    @property
    def id(self) -> str:
        """Return Room id.

        This is a numerical ID generated by SmartCocoon
        This value can change if the room is re-added to
        an cloud account
        """
        return self._id

    @property
    def name(self) -> str:
        """Return Room name"""
        return self._name
